@@ -1,131 +1,27 @@
 import base64
-import inspect
 import json
-import logging
 import warnings
 from collections import defaultdict, namedtuple
 from enum import Enum, EnumMeta
 from types import FunctionType
-from typing import Union, get_type_hints
+from typing import Any
 
-from decouple import config
-
-from lpipe import kinesis, sentry, sqs
-from lpipe.exceptions import (
-    FailButContinue,
-    FailCatastrophically,
-    InvalidConfigurationError,
-    InvalidPathError,
-    InvalidPayloadError,
-    LpipeBaseException,
-)
-from lpipe.logging import ServerlessLogger
-from lpipe.utils import AutoEncoder, _repr, exception_to_str, get_enum_value, get_nested
+import lpipe.exceptions
+import lpipe.logging
+from lpipe import normalize, signature, utils
+from lpipe.action import Action
+from lpipe.contrib import kinesis, mindictive, sqs
+from lpipe.payload import Payload
+from lpipe.queue import Queue, QueueType
 
 
-class QueueType(Enum):
-    RAW = 1
-    KINESIS = 2
-    SQS = 3
-
-
-class Queue:
-    """Represents a queue path in an Action.
-
-    Note:
-        Kinesis uses name.
-        SQS uses name or url.
-
-    Args:
-        type (QueueType)
-        path (str): Value of the "path" field set on the message we'll send to this queue.
-        name (str, optional): Queue name
-        url (str, optional): Queue URL/URI
-
-    Attributes:
-        type (QueueType)
-        path (str)
-        name (str)
-        url (str)
-
-    """
-
-    def __init__(self, type, path, name=None, url=None):
-        assert name or url
-        assert isinstance(type, QueueType)
-        self.type = type
-        self.path = path
-        self.name = name
-        self.url = url
-
-    def __repr__(self):
-        return _repr(self, ["type", "name", "url"])
-
-
-def clean_path(path_enum: EnumMeta, path):
-    if isinstance(path, Queue):
-        return path
-    else:
-        try:
-            return get_enum_value(path_enum, path)
-        except Exception as e:
-            raise InvalidPathError(
-                "Unable to cast your path identifier to an enum."
-            ) from e
-
-
-class Action:
-    def __init__(
-        self, functions=[], paths=[], required_params=None, include_all_params=False
-    ):
-        assert functions or paths
-        self.functions = functions
-        self.paths = paths
-        self.required_params = required_params
-        self.include_all_params = include_all_params
-
-    def __repr__(self):
-        return _repr(self, ["functions", "paths"])
-
-    def copy(self):
-        return type(self)(
-            functions=self.functions,
-            paths=[
-                p if isinstance(p, Queue) else str(p).split(".")[-1] for p in self.paths
-            ],
-            required_params=self.required_params,
-        )
-
-
-class Payload:
-    def __init__(self, path, kwargs: dict, event_source=None):
-        self.path = path
-        self.kwargs = kwargs
-        self.event_source = event_source
-
-    def validate(self, path_enum: EnumMeta = None):
-        if path_enum:
-            assert isinstance(
-                clean_path(path_enum, self.path), path_enum
-            ) or isinstance(self.path, Queue)
-        else:
-            assert isinstance(self.path, Queue)
-        return self
-
-    def to_dict(self) -> dict:
-        return {"path": self.path, "kwargs": self.kwargs}
-
-    def __repr__(self):
-        return _repr(self, ["path", "kwargs"])
-
-
-def build_response(n_records, n_ok, logger) -> dict:
+def build_event_response(n_records, n_ok, logger) -> dict:
     response = {
         "event": "Finished.",
         "stats": {"received": n_records, "successes": n_ok},
     }
     if hasattr(logger, "events") and logger.events:
-        response["logs"] = json.dumps(logger.events, cls=AutoEncoder)
+        response["logs"] = json.dumps(logger.events, cls=utils.AutoEncoder)
     return response
 
 
@@ -137,8 +33,9 @@ def process_event(
     path_enum: EnumMeta = None,
     default_path=None,
     logger=None,
-    debug=False,
-):
+    debug: bool = False,
+    exception_handler: FunctionType = None,
+) -> dict:
     """Process an AWS Lambda event.
 
     Args:
@@ -150,44 +47,30 @@ def process_event(
         default_path: A string or Enum which will be run for every message received.
         logger:
         debug (bool):
+        exception_handler (FunctionType): A function which will be used to capture exceptions (e.g. contrib.sentry.capture)
     """
-    try:
-        if not logger:
-            logger = ServerlessLogger(
-                level=logging.DEBUG if debug else logging.INFO,
-                process=getattr(
-                    context, "function_name", config("FUNCTION_NAME", default=None)
-                ),
-            )
-
-        if debug and isinstance(logger, ServerlessLogger):
-            logger.persist = True
-    except Exception as e:
-        raise InvalidConfigurationError("Failed to initialize logger.") from e
-
+    logger = lpipe.logging.setup(logger=logger, context=context, debug=debug)
     logger.debug(f"Event received. queue: {queue_type}, event: {event}")
+
     try:
         assert isinstance(queue_type, QueueType)
     except AssertionError as e:
-        raise InvalidConfigurationError(f"Invalid queue type '{queue_type}'") from e
+        raise lpipe.exceptions.InvalidConfigurationError(
+            f"Invalid queue type '{queue_type}'"
+        ) from e
 
-    if not path_enum:
-        try:
-            path_enum = Enum("AutoPath", [k.upper() for k in paths.keys()])
-            paths = {clean_path(path_enum, k): v for k, v in paths.items()}
-        except KeyError as e:
-            raise InvalidConfigurationError from e
+    paths, path_enum = normalize.normalize_path_enum(path_enum=path_enum, paths=paths)
 
     successful_records = []
     records = get_records_from_event(queue_type, event)
     try:
         assert isinstance(records, list)
     except AssertionError as e:
-        logger.error(f"'records' is not a list {exception_to_str(e)}")
-        return build_response(0, 0, logger)
+        logger.error(f"'records' is not a list {utils.exception_to_str(e)}")
+        return build_event_response(0, 0, logger)
 
-    output = []
-    exceptions = []
+    _output = []
+    _exceptions = []
     for encoded_record in records:
         ret = None
         try:
@@ -202,16 +85,16 @@ def process_event(
                 _event_source = get_event_source(queue_type, encoded_record)
                 payload = Payload(_path, _kwargs, _event_source).validate(path_enum)
             except AssertionError as e:
-                raise InvalidPayloadError(
+                raise lpipe.exceptions.InvalidPayloadError(
                     "'path' or 'kwargs' missing from payload."
                 ) from e
             except TypeError as e:
-                raise InvalidPayloadError(
-                    f"Bad record provided for queue type {queue_type}. {encoded_record} {exception_to_str(e)}"
+                raise lpipe.exceptions.InvalidPayloadError(
+                    f"Bad record provided for queue type {queue_type}. {encoded_record} {utils.exception_to_str(e)}"
                 ) from e
 
             with logger.context(bind={"payload": payload.to_dict()}):
-                logger.log(f"Record received.")
+                logger.log("Record received.")
 
             # Run your path/action/functions against the payload found in this record.
             ret = execute_payload(
@@ -224,39 +107,41 @@ def process_event(
                 debug=debug,
             )
 
-            # Will handle any cleanup necessary for a successful record later.
+            # Will handle cleanup for successful records later, if necessary.
             successful_records.append(encoded_record)
-        except FailButContinue as e:
+        except lpipe.exceptions.FailButContinue as e:
             # CAPTURES:
-            #    InvalidPayloadError
-            #    InvalidPathError
+            #    lpipe.exceptions.InvalidPayloadError
+            #    lpipe.exceptions.InvalidPathError
             logger.error(str(e))
-            sentry.capture(e)
+            if exception_handler:
+                exception_handler(e)
             continue  # User can say "bad thing happened but keep going." This drops poisoned records on the floor.
-        except FailCatastrophically as e:
+        except lpipe.exceptions.FailCatastrophically as e:
             # CAPTURES:
-            #    InvalidConfigurationError
+            #    lpipe.exceptions.InvalidConfigurationError
             # raise (later)
-            exceptions.append({"exception": e, "record": encoded_record})
-        output.append(ret)
+            _exceptions.append({"exception": e, "record": encoded_record})
+        _output.append(ret)
 
-    response = build_response(
+    response = build_event_response(
         n_records=len(records), n_ok=len(successful_records), logger=logger
     )
 
-    if exceptions:
-        # Handle any cleanup necessary for successful records before creating an error state.
+    if _exceptions:
+        # Handle cleanup for successful records, if necessary, before creating an error state.
         advanced_cleanup(queue_type, successful_records, logger)
 
         logger.info(
             f"Encountered exceptions while handling one or more records. RESPONSE: {response}"
         )
-        raise FailCatastrophically(exceptions)
+        raise lpipe.exceptions.FailCatastrophically(_exceptions)
 
-    if any(output):
-        response["output"] = output
+    if any(_output):
+        response["output"] = _output
     if debug:
-        response["debug"] = json.dumps({"records": records}, cls=AutoEncoder)
+        response["debug"] = json.dumps({"records": records}, cls=utils.AutoEncoder)
+
     return response
 
 
@@ -267,8 +152,9 @@ def execute_payload(
     logger,
     event,
     context,
-    debug=False,
-):
+    debug: bool = False,
+    exception_handler: FunctionType = None,
+) -> Any:
     """Execute functions, paths, and shortcuts in a Path.
 
     Args:
@@ -278,19 +164,19 @@ def execute_payload(
         logger:
         event: https://docs.aws.amazon.com/lambda/latest/dg/python-handler.html
         context: https://docs.aws.amazon.com/lambda/latest/dg/python-handler.html
+        debug (bool):
+        exception_handler (FunctionType): A function which will be used to capture exceptions (e.g. contrib.sentry.capture)
     """
     if not logger:
-        logger = ServerlessLogger()
+        logger = lpipe.logging.LPLogger()
 
     ret = None
 
     if not isinstance(payload.path, path_enum):
-        payload.path = clean_path(path_enum, payload.path)
+        payload.path = normalize.normalize_path(path_enum, payload.path)
 
     if isinstance(payload.path, Enum):  # PATH
-        # Allow someone to simplify their definition of a Path to a list of functions.
-        if all([isinstance(f, FunctionType) for f in paths[payload.path]]):
-            paths[payload.path] = [Action(functions=action)]
+        paths[payload.path] = normalize.normalize_actions(paths[payload.path])
 
         for action in paths[payload.path]:
             assert isinstance(action, Action)
@@ -304,8 +190,8 @@ def execute_payload(
                 for k in dummy:
                     action_kwargs.pop(k, None)
             except (TypeError, AssertionError) as e:
-                raise InvalidPayloadError(
-                    f"Failed to run {payload.path.name} {action} due to {exception_to_str(e)}"
+                raise lpipe.exceptions.InvalidPayloadError(
+                    f"Failed to run {payload.path.name} {action} due to {utils.exception_to_str(e)}"
                 ) from e
 
             # Run action functions
@@ -328,50 +214,27 @@ def execute_payload(
                                 },
                             }
                         )
-
-                    if ret:
-                        _payloads = []
-                        try:
-                            if isinstance(ret, Payload):
-                                _payloads.append(ret.validate(path_enum))
-                            elif isinstance(ret, list):
-                                for r in ret:
-                                    if isinstance(r, Payload):
-                                        _payloads.append(r.validate(path_enum))
-                        except Exception as e:
-                            logger.debug(exception_to_str(e))
-                            raise FailButContinue(
-                                f"Something went wrong while extracting Payloads from a function return value. {ret}"
-                            ) from e
-
-                        for p in _payloads:
-                            logger.debug(f"Function returned a Payload. Executing. {p}")
-                            try:
-                                ret = execute_payload(
-                                    p, path_enum, paths, logger, event, context, debug
-                                )
-                            except Exception as e:
-                                logger.debug(exception_to_str(e))
-                                raise FailButContinue(
-                                    f"Failed to execute returned Payload. {p}"
-                                ) from e
-                except LpipeBaseException:
+                    ret = return_handler(
+                        ret, path_enum, paths, logger, event, context, debug
+                    )
+                except lpipe.exceptions.LPBaseException:
                     # CAPTURES:
-                    #    FailButContinue
-                    #    FailCatastrophically
+                    #    lpipe.exceptions.FailButContinue
+                    #    lpipe.exceptions.FailCatastrophically
                     raise
                 except Exception as e:
                     logger.error(
-                        f"Skipped {payload.path.name} {f.__name__} due to unhandled Exception. This is very serious; please update your function to handle this. Reason: {exception_to_str(e)}"
+                        f"Skipped {payload.path.name} {f.__name__} due to unhandled Exception. This is very serious; please update your function to handle this. Reason: {utils.exception_to_str(e)}"
                     )
-                    sentry.capture(e)
+                    if exception_handler:
+                        exception_handler(e)
                     if debug:
-                        raise FailCatastrophically() from e
+                        raise lpipe.exceptions.FailCatastrophically() from e
 
             # Run action paths / shortcuts
             for path_descriptor in action.paths:
                 _payload = Payload(
-                    clean_path(path_enum, path_descriptor),
+                    normalize.normalize_path(path_enum, path_descriptor),
                     action_kwargs,
                     payload.event_source,
                 ).validate(path_enum)
@@ -399,7 +262,38 @@ def execute_payload(
     return ret
 
 
-def advanced_cleanup(queue_type, records, logger, **kwargs):
+def return_handler(
+    ret, path_enum: EnumMeta, paths: dict, logger, event, context, debug: bool
+) -> Any:
+    if not ret:
+        return ret
+    _payloads = []
+    try:
+        if isinstance(ret, Payload):
+            _payloads.append(ret.validate(path_enum))
+        elif isinstance(ret, list):
+            for r in ret:
+                if isinstance(r, Payload):
+                    _payloads.append(r.validate(path_enum))
+    except Exception as e:
+        logger.debug(utils.exception_to_str(e))
+        raise lpipe.exceptions.FailButContinue(
+            f"Something went wrong while extracting Payloads from a function return value. {ret}"
+        ) from e
+
+    for p in _payloads:
+        logger.debug(f"Function returned a Payload. Executing. {p}")
+        try:
+            ret = execute_payload(p, path_enum, paths, logger, event, context, debug)
+        except Exception as e:
+            logger.debug(utils.exception_to_str(e))
+            raise lpipe.exceptions.FailButContinue(
+                f"Failed to execute returned Payload. {p}"
+            ) from e
+    return ret
+
+
+def advanced_cleanup(queue_type: QueueType, records: list, logger, **kwargs):
     """If exceptions were raised, cleanup all successful records before raising.
 
     Args:
@@ -412,7 +306,7 @@ def advanced_cleanup(queue_type, records, logger, **kwargs):
     # If the queue type was not handled, no cleanup was necessary by lpipe.
 
 
-def cleanup_sqs_records(records, logger):
+def cleanup_sqs_records(records: list, logger):
     base_err_msg = (
         "Unable to delete successful records messages from SQS queue. AWS should "
         "still handle this automatically when the lambda finishes executing, but "
@@ -424,10 +318,10 @@ def cleanup_sqs_records(records, logger):
         messages = defaultdict(list)
         for record in records:
             m = Message(
-                message_id=get_nested(record, ["messageId"]),
-                receipt_handle=get_nested(record, ["receiptHandle"]),
+                message_id=mindictive.get_nested(record, ["messageId"]),
+                receipt_handle=mindictive.get_nested(record, ["receiptHandle"]),
             )
-            messages[get_nested(record, ["eventSourceARN"])].append(m)
+            messages[mindictive.get_nested(record, ["eventSourceARN"])].append(m)
         for k in messages.keys():
             queue_url = sqs.get_queue_url(k)
             sqs.batch_delete_messages(
@@ -439,10 +333,10 @@ def cleanup_sqs_records(records, logger):
             )
     except KeyError as e:
         logger.warning(
-            f"{base_err_msg} If you're testing, this is not an issue. {exception_to_str(e)}"
+            f"{base_err_msg} If you're testing, this is not an issue. {utils.exception_to_str(e)}"
         )
     except Exception as e:
-        logger.warning(f"{base_err_msg} {exception_to_str(e)}")
+        logger.warning(f"{base_err_msg} {utils.exception_to_str(e)}")
 
 
 def build_action_kwargs(action: Action, kwargs: dict) -> dict:
@@ -478,7 +372,7 @@ def build_kwargs(kwargs: dict, functions: list, required_params: list = None) ->
     """
     kwargs_union = {}
     if not required_params and functions:
-        kwargs_union = validate_signature(functions, kwargs)
+        kwargs_union = signature.validate(functions, kwargs)
     elif required_params and isinstance(required_params, list):
         for param in required_params:
             param_name = param[0] if isinstance(param, tuple) else param
@@ -486,7 +380,9 @@ def build_kwargs(kwargs: dict, functions: list, required_params: list = None) ->
                 # Assert required field was provided.
                 assert param_name in kwargs
             except AssertionError as e:
-                raise InvalidPayloadError(f"Missing param '{param_name}'") from e
+                raise lpipe.exceptions.InvalidPayloadError(
+                    f"Missing param '{param_name}'"
+                ) from e
 
             # Set param in kwargs. If the param is a tuple, use the [1] as the new key.
             if isinstance(param, tuple) and len(param) == 2:
@@ -496,97 +392,12 @@ def build_kwargs(kwargs: dict, functions: list, required_params: list = None) ->
     elif not required_params:
         return {}
     else:
-        raise InvalidPayloadError(
+        raise lpipe.exceptions.InvalidPayloadError(
             "You either didn't provide functions or required_params was not an instance of list or NoneType."
         )
     return kwargs_union
 
 
-def _merge(functions: list, iter):
-    """Get a set of attributes describing several functions.
-
-    Args:
-        functions (list): list of functions (FunctionType)
-
-    Raises:
-        TypeError: If two functions have the same parameter name with different types or defaults.
-    """
-    output = {}
-    for f in functions:
-        assert isinstance(f, FunctionType)
-        i = iter(f)
-        for k, v in i.items():
-            if k in output:
-                try:
-                    assert v == output[k]
-                except AssertionError as e:
-                    raise TypeError(
-                        f"Incompatible functions {functions}: {k} represented as both {v} and {output[k]}"
-                    ) from e
-            else:
-                output[k] = v
-    return output
-
-
-def _merge_signatures(functions: list):
-    """Create a combined list of function parameters."""
-    return _merge(functions, lambda f: inspect.signature(f).parameters)
-
-
-def _merge_type_hints(functions: list):
-    """Create a combined list of function parameter type hints."""
-    return _merge(functions, lambda f: get_type_hints(f))
-
-
-def _get_defaults(signature):
-    """Create a dict of function parameters with defaults."""
-    return {
-        k: v.default
-        for k, v in signature.items()
-        if v.default is not inspect.Parameter.empty
-    }
-
-
-def validate_signature(functions: list, params: dict) -> dict:
-    """Validate and build kwargs for a set of functions based on their signatures.
-
-    Args:
-        functions (list): functions
-        params (dict): kwargs provided in the event's message
-
-    Returns:
-        dict: validated kwargs required by the provided set of functions
-    """
-    signature = _merge_signatures(functions)
-    defaults = _get_defaults(signature)
-    hints = _merge_type_hints(functions)
-
-    validated = {}
-    for k, v in signature.items():
-        if k in params:
-            p = params[k]
-            if k in hints:
-                t = hints[k]
-                try:
-                    if hasattr(t, "__origin__") and t.__origin__ is Union:
-                        # https://stackoverflow.com/a/49471187
-                        assert any([isinstance(p, typ) for typ in t.__args__])
-                    else:
-                        assert isinstance(p, t)
-                except AssertionError as e:
-                    raise TypeError(f"Type of {k} should be {t} not {type(p)}.") from e
-                validated[k] = p
-            else:
-                validated[k] = p
-        elif k not in ("kwargs", "args"):
-            try:
-                assert k in defaults
-            except AssertionError as e:
-                raise TypeError(f"{functions} missing required argument: '{k}'") from e
-
-    return validated
-
-
 def get_raw_payload(record) -> dict:
     """Decode and validate a json record."""
     assert record is not None
@@ -616,7 +427,7 @@ def get_records_from_event(queue_type: QueueType, event):
 
 def get_event_source(queue_type: QueueType, record):
     if queue_type in (QueueType.RAW, QueueType.KINESIS, QueueType.SQS):
-        return get_nested(record, ["event_source_arn"], None)
+        return mindictive.get_nested(record, ["event_source_arn"], None)
     warnings.warn(f"Unable to fetch event_source for {queue_type} record.")
     return None
 
@@ -630,8 +441,8 @@ def get_payload_from_record(queue_type: QueueType, record, validate=True) -> dic
         if queue_type == QueueType.SQS:
             payload = get_sqs_payload(record)
     except json.JSONDecodeError as e:
-        raise InvalidPayloadError(
-            f"Payload contained invalid json. {exception_to_str(e)}"
+        raise lpipe.exceptions.InvalidPayloadError(
+            f"Payload contained invalid json. {utils.exception_to_str(e)}"
         ) from e
     if validate:
         for field in ["path", "kwargs"]:
@@ -648,4 +459,6 @@ def put_record(queue: Queue, record: dict):
         try:
             return sqs.put_message(queue_url=queue.url, data=record)
         except Exception as e:
-            raise FailCatastrophically(f"Failed to send message to {queue}") from e
+            raise lpipe.exceptions.FailCatastrophically(
+                f"Failed to send message to {queue}"
+            ) from e
